# -*- coding: utf-8 -*-
"""24Mayo2021Graph.ipynb

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/1CwpIGoYV9R8Cdt0AEw1w5gowgv7zlSri

**Problema 2**

La clase MyGraph es una implementación basada en diccionarios para representar un grafo dirigido no ponderado.

 Para simplificar el problema, supondremos que los vértices son números enteros no negativos (0,1,2,..). 
 
 Implementa la función minimumPath, que reciba dos vértices, start y end, y devuelve una lista de Python con los vértices que forman el camino mínimo desde start a end, ambos inclusive. La función minimumPath debe utilizar el algoritmo de camino mínimo de Dijkstra. Como el grafo es un grafo no ponderado, se considera como camino mínimo aquel que tenga un menor número de aristas. Está permitido utilizar estructuras de Python como las listas o los diccionarios.
"""

import math
class MyGraph():
    
    def __init__(self,n):
       """Crea un grafo con n vértices (0,1,...,n-1)"""
       self.vertices={}
       for i in range(n):
           self.vertices[i]=[]
   
    def addConnection(self,i,j):
        if i not in range(len(self.vertices)):
           print(i, " is not a vertex!!!")
           return
        if j not in range(len(self.vertices)):
           print(j, " is not a vertex!!!")
           return
        self.vertices[i].append(j)
 

    def minDistance(self, distances, visited): 
        """This functions returns the vertex (index) whose associated value in
        the dictionary distances is the smallest value. We 
        only consider the set of vertices that have not been visited"""
        # Initilaize minimum distance for next node 
        min = math.inf 

        #returns the vertex with minimum distance from the non-visited vertices
        for vertex in self.vertices.keys(): 
            if distances[vertex] <= min and visited[vertex] == False: 
                min = distances[vertex] #update the new smallest
                min_vertex = vertex      #update the index of the smallest
    
        return min_vertex 




    def dijkstra(self, origin): 
        """"This function takes a vertex v and calculates its mininum path 
        to the rest of vertices by using the Dijkstra algoritm"""  
        
        visited={} #visisted is a dictionary whose keys are the verticies of our graph. 
        previous={} # #this dictionary will save the previous vertex for the key in the minimum path
        distances={} #This dictionary will save the accumulate distance from the  origin to the vertex (key)
        
        for v in self.vertices.keys():
            visited[v]=False
            previous[v]=None
            distances[v]=math.inf


        #The distance from origin to itself is 0
        distances[origin] = 0

        for n in range(len(self.vertices)): 
            # Pick the vertex with the minimum distance vertex.
            # u is always equal to origin in first iteration 
            u = self.minDistance(distances, visited) 

            visited[u] = True
            
            #we must visit all adjacent vertices (neighbours) for u
            for v in self.vertices[u]: 
                if visited[v]==False and distances[v]>distances[u]+1:
                    #we must update because its distance is greater than the new distance
                    distances[v]=distances[u]+1   
                    previous[v]=u       
                
        #finally, we print the minimum path from origin to the other vertices
        return distances,previous

    def minimumPath(self, start, end):
        """returns the shortest path from start to end and its distance"""
        print("Mininum path from ",start, " to ", end)
        
        if start not in range(len(self.vertices)):
           print(start, " is not a vertex!!!")
           return None, None

        if end not in range(len(self.vertices)):
           print(end, " is not a vertex!!!")
           return None, None

            
        distances,previous=self.dijkstra(start)
        if distances[end]==math.inf:
            print("There is not path from {} to {}".format(start,end))
            return None, math.inf

        minimum_path=[]
        prev=previous[end]
        while prev!=None:
            minimum_path.insert(0,prev)
            prev=previous[prev]

        minimum_path.append(end) 

        return minimum_path, distances[end]
  
    def __str__(self):
        result=''
        for u in self.vertices.keys():
            result+='\n'+str(u)+' \t: '
            for v in self.vertices[u]:
                result+=str(v)+ ','
            result=result[0:-1]
        
        return result

"""Now, we use the implementation to represent this directed and unweighted graph:


<img src='https://www.researchgate.net/publication/319327358/figure/fig8/AS:631663130337280@1527611634108/Figura-52-Ejemplo-de-Grafo-Dirigido-No-Ponderado-Extraido-de-Fre10.png'/>
"""

g=MyGraph(8)
#Faltaría el vértice 0 no conectado

#Now, we add the edges
g.addConnection(1,2) 
g.addConnection(1,3) 
g.addConnection(1,4)

g.addConnection(2,4)
g.addConnection(2,5)

g.addConnection(3,6)

g.addConnection(4,3)
g.addConnection(4,6)
g.addConnection(4,7)

g.addConnection(5,4)
g.addConnection(5,7)

g.addConnection(7,6)
print(g)


path,d=g.minimumPath(0,1)
print(path,d)


path,d=g.minimumPath(2,7)
print(path,d)

"""<img src='https://www.researchgate.net/profile/Wladimir-Tavares/publication/318252493/figure/fig8/AS:668258659225604@1536336688129/Figura-31-Grafo-ponderado-direcionado-G-r-onde-r-1-2-3-4-5-6.png'>

"""

g=MyGraph(7)
#Faltaría el vértice 0 no conectado

#Now, we add the edges
g.addConnection(1,2) 
g.addConnection(1,5)

g.addConnection(2,3)

g.addConnection(3,4)
g.addConnection(3,6)

g.addConnection(4,5)
g.addConnection(4,6)


g.addConnection(5,6)

print(g)


path,d=g.minimumPath(0,1)
print(path,d)


path,d=g.minimumPath(1,4)
print(path,d)

path,d=g.minimumPath(1,6)
print(path,d)

path,d=g.minimumPath(3,6)
print(path,d)